/**
 * Copyright (c) 2004-2007 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package de.inovat.kex.bast.common.objectwrapper;


/*
 * History:
 * 2007-09-27 [KS] Code-Review, JavaDoc
 * 2007-07-09 [KS] Zugriff auf zentrale statische Konfigurationskonstanten in
 *                 ConfigurationConstants.*
 *
 * Revision 1.9  2007/11/26 12:10:19  Trier
 * <Kein Kommentar eingegeben>
 *
 */


import java.util.HashMap;
import java.util.List;
import java.util.TreeSet;
import java.util.WeakHashMap;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.ConfigurationObjectType;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.inovat.kex.bast.common.*;

/**
 * <p>Kapselt den Datenmodell-Typ "typ.de" (Daten-Endgert nach TLS).</p>
 * <pre>
 * deLve     Datenendgertekanal der Funktionsgruppe 1, LokaleVerkehrsErfassung.
 * deAxl     Datenendgertekanal der Funktionsgruppe 2, AchsLastdaten.
 * deUfd     Datenendgertekanal der Funktionsgruppe 3, UmFeldDaten.
 * deWzg     Datenendgertekanal der Funktionsgruppe 4, WechselZeichenGeber.
 * deVlt     Datenendgertekanal der Funktionsgruppe 6, Netze der VerkehrLeitTechnik.
 * deSys     Datenendgertekanal der Funktionsgruppe 254, System.
 * </pre>
 *
 * @author LOGOS GmbH Rostock (www.logos-ib.de)
 * @author Ralf Trier
 * @version $Id: DEKanal.java 442 2008-11-10 06:59:32Z Schawaller $
 */
public abstract class DEKanal extends ConfigurationObjectWrapper {

  /** DebugLogger fuer Debug-Ausgaben. */
  private static Debug debug = Debug.getLogger();

  private static WeakHashMap<String,DEKanal> __Cache = new WeakHashMap<String,DEKanal>();
  static HashMap<String,DEKanal.Lve> __DeLveCache = new HashMap<String,DEKanal.Lve>();
  private static DataDescription __DataDescrFehlerMeldung;
  private static ConfigurationObjectType __ConfigurationObjectTypeDeLve;
  private static ConfigurationObjectType __ConfigurationObjectTypeDeAxl;
  private static ConfigurationObjectType __ConfigurationObjectTypeDeUfd;
  private static ConfigurationObjectType __ConfigurationObjectTypeDeWzg;
  private static ConfigurationObjectType __ConfigurationObjectTypeDeVlt;
  private static ConfigurationObjectType __ConfigurationObjectTypeDeSys;
  private static AttributeGroup __AtgDE;

  private String _Bezeichnung;
  boolean _Cluster;
  int _EAKanal;
  int _DEKanal;
  String _UmsetzungsModul;
  private EAK _Eak;

  /**
   * <p>Setzt den EAK, an den der DEKanal angeschlossen ist.</p>
   *
   * @param eak EAK, an den der DEKanal angeschlossen ist.
   */
  public void setEAK(EAK eak) {
    _Eak = eak;
  }

  /**
   * <p>Gibt den EAK zur&uuml;ck, an den der DEKanal angeschlossen ist
   * (L&auml;dt automatisch die TLS-Struktur durch
   * {@link de.inovat.kex.bast.common.objectwrapper.KRI#getAll()}).</p>
   *
   * @return EAK
   */
  public EAK getEAK() {
    if (_Eak == null) {
      try {
        KRI.getAll();
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
    }
    return _Eak;
  }

  /**
   * <p>Gibt das zust&auml;ndige UmsetzungsModul zur&uuml;ck.</p>
   *
   * @return String
   */
  public String getUmsetzungsModul() {
    return _UmsetzungsModul;
  }


  /**
   * <p>Gibt die DataDescription zur&uuml;ck, die f&uuml;r die "Erg&auml;nzende
   * DE-Fehlermeldung" zust&auml;ndig ist.</p>
   *
   * @return String
   */
  public abstract DataDescription getDataDescrErgaenzendeFehlerMeldung();

  /**
   * <p>Gibt die {@link DataDescription} zur&uuml;ck, die f&uuml;r die
   * "Negative Quittung" zust&auml;ndig ist.</p>
   *
   * @return String
   */
  public abstract DataDescription getDataDescrNegativeQuittung();

  /**
   * <p>Gibt die {@link DataDescription} zur&uuml;ck, die f&uuml;r die
   * "Negative Quittung"  zust&auml;ndig ist.</p>
   *
   * @return String
   */
  public DataDescription getDataDescrTLSFehlerMeldung() {
    if (__DataDescrFehlerMeldung==null) {
      DataModel model = NetzClientDavConnection.getInstance().getDataModel();
      AttributeGroup atg = model.getAttributeGroup("atg.tlsGloDeFehler");
      Aspect asp = model.getAspect("asp.tlsAntwort");
      __DataDescrFehlerMeldung = new DataDescription(atg, asp, (short)0);
    }
    return __DataDescrFehlerMeldung;
  }

  /**
   * <p>Gibt die Bezeichnung zur&uuml;ck.</p>
   *
   * @return String
   */
  public String getBezeichnung() {
    return _Bezeichnung;
  }

  /**
   * <p>Gibt an, ob es sich bei der DE um einen Clusterkanal handelt.</p>
   *
   * @return boolean
   */
  public boolean isCluster() {
    return _Cluster;
  }

  /**
   * <p>Gibt an, ob es sich bei der DE um einen SteuerKanal (deSys 254)
   * handelt.</p>
   *
   * @return boolean
   */
  public boolean isSteuerKanal() {
    return (this instanceof DEKanal.Sys);
  }

  /**
   * <p>Gibt die DE-Nummer des Datenendgertes zur&uuml;ck.</p>
   *
   * @return byte
   */
  public int getDEKanal() {
    return _DEKanal;
  }

  /**
   * <p>Gibt die zugeordnete EA-Kanalnummer des DE am zugehrigen EAK
   * zur&uuml;ck.</p>
   *
   * @return byte
   */
  public int getEAKanal() {
    return _EAKanal;
  }

  /**
   * <p>Erzeugt das Wrapper-Object f&uuml;r das Konfigurationsobjekt mit der
   * angegebenen Pid</p>
   *
   * @see ConfigurationObjectWrapper#ConfigurationObjectWrapper(String pid)
   *
   * @param pid String
   * @throws DataModellMismatchException
   * @throws DataModellObjectNotFoundException
   */
  private DEKanal(String pid) throws DataModellMismatchException, DataModellObjectNotFoundException {
    super(pid);
  }

  /**
   * <p>Erzeugt das Wrapper-Object fr das angegebene Konfigurationsobject.</p>
   *
   * @param confObj ConfigurationObject
   * @throws DataModellMismatchException
   */
  private DEKanal(ConfigurationObject confObj) throws DataModellMismatchException {
    super(confObj);
  }

  /**
   * <p>Prft und Initialisierung das Wrapper-Objekt anhand der Daten des
   * angegebenen Konfigurationsobjekts.</p>
   *
   * @param deKanal ConfigurationObject
   * @throws DataModellMismatchException
   */
  protected void initData(ConfigurationObject deKanal) throws DataModellMismatchException {
    Data oData = deKanal.getConfigurationData(__AtgDE);
    if (oData != null) {
      _Bezeichnung = oData.getTextValue("Bezeichnung").getText();
      // <zustand name="Ja" wert="1"/><zustand name="Nein" wert="0"/>
      _Cluster = (oData.getUnscaledValue("Cluster").intValue()==1) ? true: false;
      _EAKanal = oData.getUnscaledValue("EAKanal").intValue();
      _DEKanal = oData.getUnscaledValue("DEKanal").intValue();
      _UmsetzungsModul = oData.getTextValue("UmsetzungsModul").getText();
    }
    __Cache.put(this.getPID(), this);
  }

  private static synchronized void init() {
    if (__AtgDE == null) {
      DataModel model = NetzClientDavConnection.getInstance().getDataModel();
      __AtgDE = model.getAttributeGroup("atg.de");
      __ConfigurationObjectTypeDeLve = (ConfigurationObjectType)model.getType("typ.deLve");
      __ConfigurationObjectTypeDeAxl = (ConfigurationObjectType)model.getType("typ.deAxl");
      __ConfigurationObjectTypeDeUfd = (ConfigurationObjectType)model.getType("typ.deUfd");
      __ConfigurationObjectTypeDeWzg = (ConfigurationObjectType)model.getType("typ.deWzg");
      __ConfigurationObjectTypeDeVlt = (ConfigurationObjectType)model.getType("typ.deVlt");
      __ConfigurationObjectTypeDeSys = (ConfigurationObjectType)model.getType("typ.deSys");
      ConfigurationObjectType configurationObjectTypeDe = (ConfigurationObjectType)model.getType("typ.de");

      List<SystemObject> list = configurationObjectTypeDe.getObjects();
      model.getConfigurationData(list, __AtgDE);
    }
  }


  /**
   * <p>Liefert das Wrapper-Objekt zum angegebenen Konfigurationsobjekt.</p>
   *
   * @param confObj ConfigurationObject
   * @return DEKanal
   * @throws DataModellMismatchException
   */
  static public DEKanal getInstance(ConfigurationObject confObj) throws DataModellMismatchException {
    DEKanal deKanal = __Cache.get(confObj.getPid());
    if (deKanal == null) {
      deKanal = getInstanceIntern(confObj);
    }
    return deKanal;
  }

  /**
   * <p>Liefert das Wrapper-Objekt zum Konfigurationsobjekt mit der angegebenen
   * Pid.</p>
   *
   * @param pid String
   * @return DEKanal
   * @throws DataModellMismatchException
   */
  static public DEKanal getInstance(String pid) throws DataModellMismatchException {
    DEKanal deKanal = __Cache.get(pid);
    if (deKanal == null) {
      SystemObject o = NetzClientDavConnection.getInstance().getDataModel().getObject(pid);
      if ((o!=null) && (o instanceof ConfigurationObject)) {
        deKanal = getInstanceIntern((ConfigurationObject)o);
      }
    }
    return deKanal;
  }

  static private synchronized DEKanal getInstanceIntern(ConfigurationObject confObj) throws DataModellMismatchException {
    DEKanal deKanal = __Cache.get(confObj.getPid());
    if (deKanal==null) {
      if (__AtgDE == null) {
        init();
      }
      if (confObj.isOfType(__ConfigurationObjectTypeDeLve)) {
        deKanal = new Lve(confObj);
      }
      if (confObj.isOfType(__ConfigurationObjectTypeDeSys)) {
        deKanal = new Sys(confObj);
      }
      if (confObj.isOfType(__ConfigurationObjectTypeDeAxl)) {
        deKanal = new Axl(confObj);
      }
      if (confObj.isOfType(__ConfigurationObjectTypeDeUfd)) {
        deKanal = new Ufd(confObj);
      }
      if (confObj.isOfType(__ConfigurationObjectTypeDeWzg)) {
        deKanal = new Wzg(confObj);
      }
      if (confObj.isOfType(__ConfigurationObjectTypeDeVlt)) {
        deKanal = new Vlt(confObj);
      }
    }
    return deKanal;
  }


// ===========================================================================

  /**
   * <p>Kapselt einen DEKanal ("typ.deLve") fr die Verkehrsdatenerfassung
   * (LVE, FG1).</p>
   */
  public static class Lve extends DEKanal implements FahrstreifenQuelle {

    private static DataDescription __DataDescrNegativeQuittung;
    private static DataDescription __DataDescrErgaenzendeFehlerMeldung;

    private Fahrstreifen _Fahrstreifen;

    private Lve(String pid) throws DataModellMismatchException, DataModellObjectNotFoundException {
      super(pid);
      __DeLveCache.put(this.getPID(), this);
    }

    private Lve(ConfigurationObject confObj) throws DataModellMismatchException {
      super(confObj);
      __DeLveCache.put(this.getPID(), this);
    }

    public void setFahrstreifen(Fahrstreifen fs) {
      _Fahrstreifen = fs;
    }

    public Fahrstreifen getFahrstreifen() {
      if (_Fahrstreifen == null) {
        try {
          LangZeitZaehlStelle.getAll();
        } catch (DataModellMismatchException ex) {
          ex.printStackTrace();
        } catch (Exception ex) {
          ex.printStackTrace();
        }
      }
      return _Fahrstreifen;
    }

    public String getTreeString() {
      if (this.getDEKanal() == 255) {
        return ("DE FG 1 - Sammelkanal 255" + " (Pid: " + getPID() + ")");
      } else {
        Fahrstreifen fs = getFahrstreifen();
        if (fs != null) {
          MessQuerschnitt mq = fs.getMessQuerschnitt();
          if ((mq != null) && (mq.getAuswerteQuerschnitt() != null)) {
            return ("DE FG 1 - " + mq.getAuswerteQuerschnitt().getTreeString() + "-" + fs.getTreeString() + " (Pid: " +
                    getPID() + ")");
          } else {
            return ("DE FG 1 - " + fs.getTreeString() + " (Pid: " + getPID() + ")");
          }
        } else {
          return ("DE FG 1 - " + getBezeichnung() + " (Pid: " + getPID() + ")");
        }
      }
    }

    ConfigurationObjectType getType() {
      return DEKanal.__ConfigurationObjectTypeDeLve;
    }

    public DataDescription getDataDescrErgaenzendeFehlerMeldung() {
      if (__DataDescrErgaenzendeFehlerMeldung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsLveErgnzendeDeFehlerMeldung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrErgaenzendeFehlerMeldung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrErgaenzendeFehlerMeldung;
    }

    public DataDescription getDataDescrNegativeQuittung() {
      if (__DataDescrNegativeQuittung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsLveNegativeQuittung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrNegativeQuittung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrNegativeQuittung;
    }

  }

// ---------------------------------------------------------------------------

  /**
   * <p>Kapselt einen DEKanal ("typ.Axl") die Erfassung von Achslastdaten
   * (TLS FG2).</p>
   */
  public static class Axl extends DEKanal implements FahrstreifenQuelle {

    private static DataDescription __DataDescrNegativeQuittung;
    private static DataDescription __DataDescrErgaenzendeFehlerMeldung;

    private Fahrstreifen _Fahrstreifen;

    private Axl(String pid) throws DataModellMismatchException, DataModellObjectNotFoundException {
      super(pid);
    }

    private Axl(ConfigurationObject confObj) throws DataModellMismatchException {
      super(confObj);
    }

    public String getTreeString() {
      if (this.getDEKanal() == 255) {
        return ("DE FG 2 - Sammelkanal 255" + " Pid: " + getPID());
      } else {
        return ("DE FG 2 " + this.getBezeichnung() + " Pid: " + getPID());
      }
    }

    ConfigurationObjectType getType() {
      return DEKanal.__ConfigurationObjectTypeDeAxl;
    }

    public void setFahrstreifen(Fahrstreifen fs) {
      _Fahrstreifen = fs;
    }

    public Fahrstreifen setFahrstreifen() {
      return _Fahrstreifen;
    }

    public DataDescription getDataDescrErgaenzendeFehlerMeldung() {
      if (__DataDescrErgaenzendeFehlerMeldung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsAxlErgnzendeDeFehlerMeldung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrErgaenzendeFehlerMeldung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrErgaenzendeFehlerMeldung;
    }

    public DataDescription getDataDescrNegativeQuittung() {
      if (__DataDescrNegativeQuittung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsAxlNegativeQuittung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrNegativeQuittung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrNegativeQuittung;
    }

  }

// ---------------------------------------------------------------------------

  /**
   * <p>Kapselt einen DEKanal ("typ.deUfd") fr die Erfassung von Umfelddaten
   * (TLS-FG3).</p>
   */
  public static class Ufd extends DEKanal {

    private static DataDescription __DataDescrNegativeQuittung;
    private static DataDescription __DataDescrErgaenzendeFehlerMeldung;

    public static TreeSet<String> _TestTreeSet = new TreeSet<String>();

    FG03Type _Type;
    int _FBNr = 0;

    private Ufd(String pid) throws DataModellMismatchException, DataModellObjectNotFoundException {
      super(pid);
    }

    private Ufd(ConfigurationObject confObj) throws DataModellMismatchException {
      super(confObj);
    }

    protected void initData(ConfigurationObject deKanal) throws DataModellMismatchException {
      super.initData(deKanal);
      _Type = FG03Type.getInstance(_UmsetzungsModul);
      if ((_Type == null) && (getEAKanal() != 255)) {
        debug.info(_ConfigurationObject.getPid() + " Umsetzungsmodul: " + _UmsetzungsModul + " - kein FG03Type");
        _TestTreeSet.add(_UmsetzungsModul);
      }
    }

    public int getFBNr() {
      return _Type.isFahrstreifenSensor() ? _FBNr : -1;
    }

    public static int getFBNr(String pid) {
      int i = 3;
      while (pid.charAt(i) != '.') {
        i++;
      }
      char ch = pid.charAt(i + 1);
      if (ch == '3') {
        return 2;
      }
      if (ch == '2') {
        return 1;
      }
      return 0;
    }

    public String getTreeString() {
      if (this.getDEKanal() == 255) {
        return ("DE FG 3 - Sammelkanal 255" + " Pid: " + getPID());
      } else {
        return ("DE FG 3 " + this.getBezeichnung() + " Pid: " + getPID());
      }
    }

    public FG03Type getFG03Type() {
      return _Type;
    }

    ConfigurationObjectType getType() {
      return DEKanal.__ConfigurationObjectTypeDeUfd;
    }

    public DataDescription getDataDescrErgaenzendeFehlerMeldung() {
      if (__DataDescrErgaenzendeFehlerMeldung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsUfdErgnzendeDeFehlerMeldung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrErgaenzendeFehlerMeldung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrErgaenzendeFehlerMeldung;
    }

    public DataDescription getDataDescrNegativeQuittung() {
      if (__DataDescrNegativeQuittung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsUfdNegativeQuittung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrNegativeQuittung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrNegativeQuittung;
    }

  }

// ---------------------------------------------------------------------------

  /**
   * <p>Kapselt einen DEKanal ("typ.deWvz") fr die WechselZeichenGeber
   * (TLS-FG4).</p>
   */
  public static class Wzg extends DEKanal {

    private static DataDescription __DataDescrNegativeQuittung;
    private static DataDescription __DataDescrErgaenzendeFehlerMeldung;

    private Wzg(String pid) throws DataModellMismatchException, DataModellObjectNotFoundException {
      super(pid);
    }

    private Wzg(ConfigurationObject confObj) throws DataModellMismatchException {
      super(confObj);
    }

    public String getTreeString() {
      if (this.getDEKanal() == 255) {
        return ("DE FG 4 - Sammelkanal 255" + " Pid: " + getPID());
      } else {
        return ("DE FG 4 " + this.getBezeichnung() + " Pid: " + getPID());
      }
    }

    ConfigurationObjectType getType() {
      return DEKanal.__ConfigurationObjectTypeDeWzg;
    }

    public DataDescription getDataDescrErgaenzendeFehlerMeldung() {
      if (__DataDescrErgaenzendeFehlerMeldung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsWzgErgnzendeDeFehlerMeldung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrErgaenzendeFehlerMeldung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrErgaenzendeFehlerMeldung;
    }

    public DataDescription getDataDescrNegativeQuittung() {
      if (__DataDescrNegativeQuittung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsWzgNegativeQuittung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrNegativeQuittung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrNegativeQuittung;
    }
  }

// ---------------------------------------------------------------------------

  /**
   * <p>Kapselt einen DEKanal ("typ.deVlt") fr die VerkehrLeitTechnik
   * (TLS-FG6, eigentlich Systemberwachung).</p>
   */
  public static class Vlt extends DEKanal {

    private static DataDescription __DataDescrNegativeQuittung;
    private static DataDescription __DataDescrErgaenzendeFehlerMeldung;

    FG06Type _Type;

    private Vlt(String pid) throws DataModellMismatchException, DataModellObjectNotFoundException {
      super(pid);
    }

    private Vlt(ConfigurationObject confObj) throws DataModellMismatchException {
      super(confObj);
      _Type = FG06Type.getInstance(_UmsetzungsModul);
      if ((_Type == null) && (getEAKanal() != 255)) {
        debug.info(String.format("%s UmsetzungsModul=%s -kein FG06Type", _ConfigurationObject.getPid(), _UmsetzungsModul));
      }
    }

    public String getTreeString() {
      if (this.getDEKanal() == 255) {
        return ("DE FG 6 - Sammelkanal 255" + " Pid: " + getPID());
      } else {
        return ("DE FG 6 " + this.getBezeichnung() + " Pid: " + getPID());
      }
    }

    ConfigurationObjectType getType() {
      return DEKanal.__ConfigurationObjectTypeDeVlt;
    }

    public DataDescription getDataDescrErgaenzendeFehlerMeldung() {
      if (__DataDescrErgaenzendeFehlerMeldung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsVltErgnzendeDeFehlerMeldung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrErgaenzendeFehlerMeldung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrErgaenzendeFehlerMeldung;
    }

    public DataDescription getDataDescrNegativeQuittung() {
      if (__DataDescrNegativeQuittung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsVltNegativeQuittung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrNegativeQuittung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrNegativeQuittung;
    }

    public FG06Type getFG06Type() {
      return _Type;
    }

  }

// ---------------------------------------------------------------------------

  /**
   * <p>Kapselt einen DEKanal ("typ.deSys") fr die SystemSteuerung
   * (TLS-FG254).</p>
   */
  public static class Sys extends DEKanal {

    private static DataDescription __DataDescrNegativeQuittung;
    private static DataDescription __DataDescrErgaenzendeFehlerMeldung;

    private Sys(String pid) throws DataModellMismatchException, DataModellObjectNotFoundException {
      super(pid);
    }

    private Sys(ConfigurationObject confObj) throws DataModellMismatchException {
      super(confObj);
    }

    public String getTreeString() {
      if (this.getDEKanal() == 255) {
        return ("DE FG 254 - Sammelkanal 255" + " Pid: " + getPID());
      } else {
        return ("DE FG 254 " + this.getBezeichnung() + " Pid: " + getPID());
      }
    }

    ConfigurationObjectType getType() {
      return DEKanal.__ConfigurationObjectTypeDeSys;
    }

    public DataDescription getDataDescrErgaenzendeFehlerMeldung() {
      if (__DataDescrErgaenzendeFehlerMeldung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsSysErgnzendeDeFehlerMeldung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrErgaenzendeFehlerMeldung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrErgaenzendeFehlerMeldung;
    }

    public DataDescription getDataDescrNegativeQuittung() {
      if (__DataDescrNegativeQuittung == null) {
        DataModel model = NetzClientDavConnection.getInstance().getDataModel();
        AttributeGroup atg = model.getAttributeGroup("atg.tlsSysNegativeQuittung");
        Aspect asp = model.getAspect("asp.tlsAntwort");
        __DataDescrNegativeQuittung = new DataDescription(atg, asp, (short) 0);
      }
      return __DataDescrNegativeQuittung;
    }

  }

}


