/**
 * Copyright (c) 2004-2007 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package de.inovat.kex.bast.common.objectwrapper;

/**
 * History:
 * 2007-09-27 [KS]  Code-Review, Logger, JavaDoc
 *
 * Revision 1.2  2007/09/28 12:18:01  Strey
 * Anpassungen fr Strmeldeberwachung, Logging, JavaDoc
 */

import java.util.ArrayList;
import java.util.List;

/**
 * <p>Kapselt die Busse zwischen den KRI's und den Steuermodulen.</p>
 * <p>Die Bezeichnung des Busses wird bestimmt, indem der letzte Teil der Pid
 * des zugehrigen AnschlussPunktes interpretiert wird. Enthlt der letzte Teil
 * der Pid das Wort "Funk" bekommt der Bus die Bezeichnung <i>Funkbus</i>
 * sonst <i>"Bus" || letzterTeil</i>.</p>
 * 
 * @author LOGOS GmbH Rostock (www.logos-ib.de)
 * @author Ralf Trier
 * @version $Id: KriBus.java 442 2008-11-10 06:59:32Z Schawaller $
 */
public class KriBus {

//  private static final de.inovat.kex.bast.funclib.logger.Logger logger =
//    de.inovat.kex.bast.funclib.logger.LoggerFactory.getLoggerFactory().getLogger(KriBus.class);

  private String _Pid;
  private String _Bezeichnung;
  private List<SteuerModul> _SteuerModulList = new ArrayList<SteuerModul>();
  private boolean _IsFunkBus;

  /**
   * <p>Gibt die Bezeichnung des Busses zurck, der auf Basis der Pid des
   * zugehrigen AnschlussPunktes bestimmt wird.</p>
   *
   * @return String
   */
  public String getBezeichnung() {
    return _Bezeichnung;
  }

  /**
   * <p>Liefert <code>true</code> zurck, wenn der Bus auf Grund der Pid als
   * Funk-Bus interpretiert wurde, ansonsten <code>false</code>.</p>
   * @return boolean
   */
  public boolean isFunkBus() {
    return _IsFunkBus;
  }

  /**
   * <p>Gibt die SteuerModule zurck, die am Bus konfiguriert wurden, oder null
   * wenn keine angeschlossen sind.</p>
   *
   * @return SteuerModule.
   */
  public SteuerModul[] getSteuerModule() {
    if ((_SteuerModulList!=null) && (_SteuerModulList.size()>0)) {
      return _SteuerModulList.toArray(new SteuerModul[_SteuerModulList.size()]);
    }
    else {
      return null;
    }
  }

  /**
   * <p>Wird von der Klasse {@link de.inovat.kex.bast.common.objectwrapper.KRI}
   * benutzt, um die an den Bus angeschlossenen Steuermodule festzulegen.</p>
   *
   * @param sm SteuerModul
   */
  public void addSteuerModul(SteuerModul sm) {
    _SteuerModulList.add(sm);
    sm.setBus(this);
  }

  /**
   * <p>Erzeugt eine Instance und bestimmt die Bezeichnung des Busses.</p>
   *
   * @param pid Pid des AnschlussPunktes am Kri
   */
  protected KriBus(String pid) {
    _Pid = pid;
    String[] strings = pid.split("\\.");
    if ((strings!=null) && (strings.length>0)) {
      if (strings[strings.length - 1].toLowerCase().indexOf("funk") > -1) {
        _Bezeichnung = "Funkbus";
        _IsFunkBus = true;
      }
      else {
        _Bezeichnung = "BUS " + strings[strings.length - 1];
        _IsFunkBus = false;
      }
    }
    else {
      _Bezeichnung = pid;
      _IsFunkBus = false;
    }
  }

  /**
   * <p>Gibt die Bezeichnung des Busses zurck, die aus der Pid ermittelt
   * wurde.</p>
   * @return String
   */
  public String toString() {
    return _Bezeichnung;
  }

  /**
   * <p>Gibt die Pid des Konfigurationsobjekt zurck.</p>
   * @return String
   */
  public String getPID() {
    return _Pid;
  }

}
