/**
 * Copyright (c) 2004-2009 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package de.inovat.kex.bast.exporter;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Dataset;
import de.bsvrz.sys.funclib.debug.Debug;
import de.inovat.kex.bast.SendeBAStStatusDaten;
import de.inovat.kex.bast.common.ChannelException;
import de.inovat.kex.bast.common.objectwrapper.LangZeitZaehlStelle;

import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Locale;

//~ KLASSEN ===================================================================

/**
 * <p>Channel zur Ausgabe von Verkehrsmengen und Geschwindigkeitsdaten in das BASt-Bestandsbandformat fr
 * Geschwindigkeitsdaten in der Version 2007. </p> <p style="padding-left:4em"><i> BASt-FORMATVORGABE FR
 * GESCHWINDIGKEITSDATEN VERSION 2007</br> Dokument V6z  F1100.6606002<br> Herausgegeben von der Bundesanstalt fr
 * Straenwesen<br> 15. November 2007<br> A. Fitschen und N. Faerber </p> <p>Die zur Verarbeitung bentigten Daten
 * werden nur an ausgewhlten Zhlstellen erhoben und in einer speziellen TLS-Datenversion bereitgestellt. Da fr die
 * Geschwindigkeitsdaten keine Folgeverarbeitung definiert und implementiert ist (die Daten werden nur fr die BASt
 * erfasst), mssen die zu verarbeitenden Daten mit der Attributgruppe fr die TLS-Datenversion 24
 * atg.tlsLveErgebnisMeldungVersion24 und Aspekt asp.tlsAntwort bergeben werden. </p> <p><b>Aufbau des
 * Dateinamens</b>:<br> Der Dateiname enthlt Informationen zum Bundesland, zur Zhlstelle, zur Datenart und ber den
 * Erfassungszeitraum.<br> In der Regel besteht der Dateiname aus 11 Stellen und einem dreistelligen Suffix (z.B.:
 * NW5033v0307.dat).<br> ndert sich die Datenstruktur im Jahresverlauf z. B. durch Umstellung von zwei auf drei
 * Fahrstreifen, so ist dieses durch einen Eintrag an zwlfter und dreizehnter Stelle des Namens kenntlich zu machen z.
 * B. durch _1 ( NW5033v0308_1.dat). Alle weiteren Dateien des Jahres behalten dann diese Versionskennung bei. Die
 * Dateinamen des nachfolgenden Jahres sind wieder ohne Versionskennung zu vergeben. Der Erfassungszeitraum nach der
 * Kennung v (Datenart: Geschwindigkeiten) enthlt in den ersten beiden Stellen das Jahr und in den beiden Folgenden
 * den Monat des betreffenden Jahres. </p> <p>Bezeichnung der Dateien am Beispiel von "NW5033v0308_1.dat":
 * <pre>
 * Position     1-2         3-6          7                 8-9   10-11   12-13     . Suffix 1-3
 * Beschreibung Bundesland: Zhlstellen- Datenart          Jahr  Monats- Version
 *              Krzel      nummer       Geschwindigkeiten (yy)  Code    (optional)
 * Beispiel     NW          5033         v                 03    08      _1        . dat
 * </pre></p>
 * <p><b>Einschrnkung: </b> Mit dem <code>DataChannelBAStVV2007</code> kann nur der Standard-Dateiname erzeugt werden,
 * da historische Konfigurationsdaten nicht ausgewertet werden. Bei der bergabe an die BASt mssen die entsprechenden
 * Suffixe durch den Bearbeiter manuell angehngt werden. </p> <p> <b>MESZ/MEZ Umstellung: </b><br> An Position 7 der
 * Stundendatenstze steht die einzige Statuskennung im Datensatz. Die Statuskennung "m" ist bei der Zeitumstellung auf
 * MESZ (letzter Sonntag im Mrz) zu vergeben. Pro Richtung wird ein Datensatz eingefgt, dessen Werte auf "0" gesetzt
 * sind. Die Statuskennung "o" ist bei der Umstellung auf MEZ (letzter Sonntag im Oktober) zu vergeben. Die Werte dieser
 * dritten Stunde sind aus der doppelten dritten Stunde zu berechnen. Dazu sind die Verkehrsmengen ber beide Stunden zu
 * addieren und die Geschwindigkeitskenngren vm, svm, v15 und v85 aus den Daten beider Stunden zu ermitteln. Alle
 * anderen Stundendatenstze enthalten an Position 7 ein Leerzeichen " ". </p> <p> <b>Geschwindigkeiten 'nicht
 * ermittelbar'</b><br> Gem ergnzender Vorgabe der BASt (EMail vom 01.09.2009 an Logos) sind fehlende
 * Geschwindigkeiten bzw. implausible Mengen in den Fahrzeugklassen so zu behandeln als lgen fr eine Stunde keine
 * Daten vor, d.h. auch in diesen Fllen besteht die Datenzeile nur aus dem Zeilenheader (14 Zeichen). </p>
 *
 * @author LOGOS GmbH Rostock (www.logos-ib.de)
 * @author K. Strey
 * @version $Id: DataChannelBAStVV2007.java 2515 2009-09-04 06:55:39Z Strey $
 */
public class DataChannelBAStVV2007 extends AbstractDataChannelBASt {

	/** Kennung fr einen nicht ermittelbaren Wert (8Bit) gem TLS. */
	protected static final int KEIN_WERT_255 = 255;

	/**
	 * Kennung fr einen nicht ermittelbaren Wert (16Bit) gem TLS: "Lassen sich (z. B. aufgrund von Schleifendefekten)
	 * einzelne Werte des Blocks nicht ermitteln, ohne dass ein Totalausfall des DE vorliegt, werden die betroffenen
	 * Bytes auf den Wert 255 (65535) gesetzt.".<br> Wird fr Geschwindigkeiten und in den Geschwindigkeitsklassen
	 * verwendet.
	 */
	protected static final int KEIN_WERT_65535 = 65535;

	/** Kennung fr einen nicht ermittelbaren Wert (16Bit mit Skalierung 0.1). */
	protected static final double KEIN_WERT_6553_5 = 6553.5;

	/** DebugLogger fuer Debug-Ausgaben. */
    private static Debug debug = Debug.getLogger();

	/** Datumformat fr Datenzeilen */
	protected static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyMMdd");

	/** Soll-Anzahl Geschwindigkeitsklassen BPA */
	protected static final int VK_ANZAHL_BPA = 10;

	/** Soll-Anzahl Geschwindigkeitsklassen LVo */
	protected static final int VK_ANZAHL_LV = 16;

	/** Soll-Anzahl Geschwindigkeitsklassen SGV */
	protected static final int VK_ANZAHL_SGV = 10;

	//~ FELDER ================================================================

	/**
	 * Zhldaten der letzten (verarbeiteten) Stunde, um bei Wechsel von MESZ->MEZ die zwei 2Uhr Stunden addieren zu
	 * knnen
	 */
	protected Dataset[] _LastArrayAtTime = null;

	/** Anzahl Fahrstreifen in Gegen-Richtung (Richtung 2) */
	protected int _FSCountGgRichtung;

	/** Anzahl Fahrstreifen in In-Richtung (Richtung 1) */
	protected int _FSCountInRichtung;

	//~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

	/**
	 * erzeugt einen DataChannel, der mit einer reset-Methode und eventuell {@link AbstractDataChannelBASt#setZaehlstelle(LangZeitZaehlStelle)}
	 * initialisiert werden muss.
	 *
	 * @param bastStatusDaten Statusinformationen.
	 */
	public DataChannelBAStVV2007(SendeBAStStatusDaten bastStatusDaten) {
		super(bastStatusDaten);
	}

	@Override
	protected String constructFilename(LangZeitZaehlStelle zaehlStelle, int year, int month, int day) {
		//throw new IllegalArgumentException("Die Erstellung der BASt-Geschwindigkeitsdaten (Version 2007) ist " + "nur monatsweise definiert. Dateinamen fr freie Zeitbereiche " + "knnen nicht generiert werden.");
		return null;  //To change body of implemented methods use File | Settings | File Templates.
	}

	//~ METHODEN ==============================================================

	/**
	 * <p>Gibt den Dateinamen fr den gewhlten Zeitraum zurck.</p> <p>Der Aufruf ist in Verbindung mit dem
	 * DataChannelBAStVV2007 nur zulssig, wenn ein Monatsbereich angegeben wird.</p> <p>Gem BASt-Vorgabe enthlt der
	 * Dateiname Informationen zum Bundesland, zur Zhlstelle, zur Datenart und ber den Erfassungszeitraum. In der
	 * Regel besteht der Dateiname aus 11 Stellen und einem dreistelligen Suffix (z.B.: "NW5033v0307.dat"). </p> <p>Der
	 * Erfassungszeitraum nach der Kennung "v" (Datenart: Geschwindigkeiten) enthlt in den ersten beiden Stellen das
	 * Jahr und in den beiden Folgenden den Monat des betreffenden Jahres.</p>
	 *
	 * @return {@link String} vollstndig klassifizierter Dateiname
	 *
	 * @throws IllegalArgumentException wird ausgelst, wenn kein Monatsbereich angegeben wird (s.o.)
	 *                                  <p/>
	 *                                  <p/>
	 *                                  Die Methode wird um die Erzeugung eines Filenamen ohne path erweitert. Diese
	 *                                  Erweiterung wird bentigt, wenn die Datei anstelle des lokalen
	 *                                  Verzeichnissystems auf den FTP-Server in einem spezfisichen Verzeichnis abgelegt
	 *                                  werden soll. Dagmar Rostek (DR) inovat.
	 */
	@Override
	protected String constructFilename(LangZeitZaehlStelle zaehlStelle, int year, int month) {
		StringBuilder sb;
		sb = new StringBuilder("");

		sb.append(zaehlStelle.getBundesland().getAbbreviation());
		sb.append(zaehlStelle.getBastNummerAsString());
		sb.append('v');
		sb.append(TWO_DIGIT_FORMAT.format(year % 100));
		sb.append(TWO_DIGIT_FORMAT.format(month + 1));

		if (FTPServerConnection.getInstance().isAenderungDatenstruktur()) {
			sb.append("_1");
		}

		sb.append(".dat");

		debug.finer(String.format("FileName: %s", sb));
		return sb.toString();

	}

	/**
	 * <p>Extrahiert die Werte fr die angegebene Richtung aus den Dataset und baut einen String mit den Werten fr die
	 * Ausgabe in die Datei zusammen. Der String ist ab Position 15 hinter Datum+Kennung+Richtungskennung in die
	 * Datenzeile einzufgen (er beginnt mit einem Leerzeichen)!</p> <p>Die Unterscheidung nach Richtungen kommt daher,
	 * dass fr jede einzelne Stunde genau zwei Datenstze (Zeilen) erzeugt werden mssen.</p> <p>Fr den Sonderfall bei
	 * der Sommerzeitumstellung im Herbst (MESZ->MEZ), mssen die zwei 2 Uhr Stundendatenstze zusammengefasst werden.
	 * Dazu knnen durch rsArrayMEZ[] die zustzlichen Daten angegeben werden. Fr das Ergebnis sind die Verkehrsmengen
	 * ber beide Stunden zu addieren und die Geschwindigkeitskenngren vm, svm, v15 und v85 aus den Daten beider
	 * Stunden zu ermitteln.</p>
	 *
	 * @param rsArray      auszugebende Intervalldaten (ATG fr TLS-Datenverion 24)
	 * @param rsArrayMEZ   optionale Angabe fr Intervalldaten, die mit denen aus <code>rsArray</code> zusammenzufassen
	 *                     sind (ist fr die 2 Stunde bei der Sommerzeitumstellung MESZ->MEZ im Oktober vorgesehen)
	 * @param isInRichtung Nummer der Fahrtrichtung der die Daten zugeordnet sind (1 = IN-Richtung, 2 = GEGEN-Richtung)
	 *
	 * @return String mit den jeweils rechtsbndig formatierten Werten
	 */
	protected String extractData(Dataset[] rsArray, Dataset[] rsArrayMEZ, boolean isInRichtung) {

		// auszugebende Data-Records ermitteln
		int fsCount = isInRichtung ? _FSCountInRichtung : _FSCountGgRichtung;
		int fsOffset = isInRichtung ? 0 : _FSCountInRichtung;
		StringBuilder sb = new StringBuilder(fsCount * 312);
		Data data[] = new Data[fsCount];
		Data dataMEZ[] = new Data[fsCount];  // optional MESZ->MEZ Doppeldatensatz

		if (rsArray != null) {
			for (int fs = 0; fs < fsCount; fs++) {
				if (rsArray[fs + fsOffset] != null) {
					data[fs] = rsArray[fs + fsOffset].getData();

					if ((rsArrayMEZ != null) && (rsArrayMEZ[fs + fsOffset] != null)) {
						dataMEZ[fs] = rsArrayMEZ[fs + fsOffset].getData();
					}
				}
			}
		}

		// wegen der TLS-ATG kann der Zugriff direkt ber getItem(...) erfolgen
		for (int fs = 0; fs < fsCount; fs++) {
			if ((data != null) && (data[fs] != null)) {

				// FS-Daten auslesen
				int qKfzNk = data[fs].getUnscaledValue("qKfzNk").intValue();
				int qKrad = data[fs].getUnscaledValue("qKrad").intValue();
				int qPkw = data[fs].getUnscaledValue("qPkw").intValue();
				int qLfw = data[fs].getUnscaledValue("qLfw").intValue();
				int qPkwA = data[fs].getUnscaledValue("qPkwA").intValue();
				int qLkw = data[fs].getUnscaledValue("qLkw").intValue();
				int qLkwA = data[fs].getUnscaledValue("qLkwA").intValue();
				int qSKfz = data[fs].getUnscaledValue("qSattelKfz").intValue();
				int qBus = data[fs].getUnscaledValue("qBus").intValue();
				double vLVo = getScaledValueAsDouble(data[fs], "vLVo");
				double vSGV = getScaledValueAsDouble(data[fs], "vSGV");
				double vBPA = getScaledValueAsDouble(data[fs], "vBPA");
				double sLVo = getScaledValueAsDouble(data[fs], "sLVo");
				double sSGV = getScaledValueAsDouble(data[fs], "sSGV");
				double sBPA = getScaledValueAsDouble(data[fs], "sBPA");
				double v85LVo = getScaledValueAsDouble(data[fs], "v85LVo");
				double v85SGV = getScaledValueAsDouble(data[fs], "v85SGV");
				double v85BPA = getScaledValueAsDouble(data[fs], "v85BPA");
				double v15LVo = getScaledValueAsDouble(data[fs], "v15LVo");
				double v15SGV = getScaledValueAsDouble(data[fs], "v15SGV");
				double v15BPA = getScaledValueAsDouble(data[fs], "v15BPA");
				int countLVo = data[fs].getUnscaledArray("qLVoGeschwKlasse").getLength();
				int vkLVo[] = new int[countLVo];

				for (int i = 0; i < countLVo; i++) {
					vkLVo[i] = data[fs].getUnscaledArray("qLVoGeschwKlasse").getValue(i).intValue();
				}

				int countSGV = data[fs].getUnscaledArray("qSGVGeschwKlasse").getLength();
				int vkSGV[] = new int[countSGV];

				for (int i = 0; i < countSGV; i++) {
					vkSGV[i] = data[fs].getUnscaledArray("qSGVGeschwKlasse").getValue(i).intValue();
				}

				int countBPA = data[fs].getUnscaledArray("qBPAGeschwKlasse").getLength();
				int vkBPA[] = new int[countBPA];

				for (int i = 0; i < countBPA; i++) {
					vkBPA[i] = data[fs].getUnscaledArray("qBPAGeschwKlasse").getValue(i).intValue();
				}

				// Spezialfall MESZ->MEZ (optionaler Datensatz angegeben?)
				if ((dataMEZ != null) && (dataMEZ[fs] != null)) {
					qKfzNk += dataMEZ[fs].getUnscaledValue("qKfzNk").intValue();
					qKrad += dataMEZ[fs].getUnscaledValue("qKrad").intValue();
					qPkw += dataMEZ[fs].getUnscaledValue("qPkw").intValue();
					qLfw += dataMEZ[fs].getUnscaledValue("qLfw").intValue();
					qPkwA += dataMEZ[fs].getUnscaledValue("qPkwA").intValue();
					qLkw += dataMEZ[fs].getUnscaledValue("qLkw").intValue();
					qLkwA += dataMEZ[fs].getUnscaledValue("qLkwA").intValue();
					qSKfz += dataMEZ[fs].getUnscaledValue("qSattelKfz").intValue();
					qBus += dataMEZ[fs].getUnscaledValue("qBus").intValue();

					double vLVoMEZ = getScaledValueAsDouble(dataMEZ[fs], "vLVo");
					double vSGVMEZ = getScaledValueAsDouble(dataMEZ[fs], "vSGV");
					double vBPAMEZ = getScaledValueAsDouble(dataMEZ[fs], "vBPA");
					double sLVoMEZ = getScaledValueAsDouble(dataMEZ[fs], "sLVo");
					double sSGVMEZ = getScaledValueAsDouble(dataMEZ[fs], "sSGV");
					double sBPAMEZ = getScaledValueAsDouble(dataMEZ[fs], "sBPA");
					double v85LVoMEZ = getScaledValueAsDouble(dataMEZ[fs], "v85LVo");
					double v85SGVMEZ = getScaledValueAsDouble(dataMEZ[fs], "v85SGV");
					double v85BPAMEZ = getScaledValueAsDouble(dataMEZ[fs], "v85BPA");
					double v15LVoMEZ = getScaledValueAsDouble(dataMEZ[fs], "v15LVo");
					double v15SGVMEZ = getScaledValueAsDouble(dataMEZ[fs], "v15SGV");
					double v15BPAMEZ = getScaledValueAsDouble(dataMEZ[fs], "v15BPA");

					vLVo = (vLVo != Double.NaN) && (vLVoMEZ != Double.NaN) ? 0.5 * (vLVo + vLVoMEZ) : Double.NaN;
					vSGV = (vSGV != Double.NaN) && (vSGVMEZ != Double.NaN) ? 0.5 * (vSGV + vSGVMEZ) : Double.NaN;
					vBPA = (vBPA != Double.NaN) && (vBPAMEZ != Double.NaN) ? 0.5 * (vBPA + vBPAMEZ) : Double.NaN;
					sLVo = (sLVo != Double.NaN) && (sLVoMEZ != Double.NaN) ? 0.5 * (sLVo + sLVoMEZ) : Double.NaN;
					sSGV = (sSGV != Double.NaN) && (sSGVMEZ != Double.NaN) ? 0.5 * (sSGV + sSGVMEZ) : Double.NaN;
					sBPA = (sBPA != Double.NaN) && (sBPAMEZ != Double.NaN) ? 0.5 * (sBPA + sBPAMEZ) : Double.NaN;
					v85LVo = (v85LVo != Double.NaN) && (v85LVoMEZ != Double.NaN) ? 0.5 * (v85LVo + v85LVoMEZ) : Double.NaN;
					v85SGV = (v85SGV != Double.NaN) && (v85SGVMEZ != Double.NaN) ? 0.5 * (v85SGV + v85SGVMEZ) : Double.NaN;
					v85BPA = (v85BPA != Double.NaN) && (v85BPAMEZ != Double.NaN) ? 0.5 * (v85BPA + v85BPAMEZ) : Double.NaN;
					v15LVo = (v15LVo != Double.NaN) && (v15LVoMEZ != Double.NaN) ? 0.5 * (v15LVo + v15LVoMEZ) : Double.NaN;
					v15SGV = (v15SGV != Double.NaN) && (v15SGVMEZ != Double.NaN) ? 0.5 * (v15SGV + v15SGVMEZ) : Double.NaN;
					v15BPA = (v15BPA != Double.NaN) && (v15BPAMEZ != Double.NaN) ? 0.5 * (v15BPA + v15BPAMEZ) : Double.NaN;

					if (countLVo == dataMEZ[fs].getUnscaledArray("qLVoGeschwKlasse").getLength()) {
						for (int i = 0; i < countLVo; i++) {
							vkLVo[i] += dataMEZ[fs].getUnscaledArray("qLVoGeschwKlasse").getValue(i).intValue();
						}
					}

					if (countSGV == data[fs].getUnscaledArray("qSGVGeschwKlasse").getLength()) {
						for (int i = 0; i < countSGV; i++) {
							vkSGV[i] += dataMEZ[fs].getUnscaledArray("qSGVGeschwKlasse").getValue(i).intValue();
						}
					}

					if (countBPA == data[fs].getUnscaledArray("qBPAGeschwKlasse").getLength()) {
						for (int i = 0; i < countBPA; i++) {
							vkBPA[i] += dataMEZ[fs].getUnscaledArray("qBPAGeschwKlasse").getValue(i).intValue();
						}
					}
				}

				// Bei der Interpretation von Geschwindigkeitsauswertungen ist der
				// Schwerverkehrsanteil von Bedeutung. Daher wird die Anzahl der
				// Fahrzeuge des Schwerverkehrs (SV) erfasst, die sich aus den
				// Fahrzeugarten des schweren Gterverkehrs (Lkw, LkwA, Sattel-Kfz) und
				// den Bussen zusammensetzt.
				// => SV Wert lsst sich aus der Klassifikation 8+1 berechnen
				// ebenso die Mengen fr LVo, SGV und BPA
				int qSV = qLkw + qLkwA + qSKfz + qBus;
				int qLVo = qKrad + qPkw + qLfw;
				int qSGV = qLkw + qLkwA + qSKfz;
				int qBPA = qPkwA + qBus;

				// Vergleichswerte Summe Fahrzeugklassen
				int qLVoVK = 0;
				int qSGVVK = 0;
				int qBPAVK = 0;

				// Ausgabe (Geschwindigkeitsklassendaten werden nur ausgegeben,
				// wenn die Anzahl V-Klassen der Formatvorgabe entspricht)
				sb.append(trimQ(qSV));

				// LVo-Gruppe
				sb.append(trimQ(qLVo));
				sb.append(trimVmStdabw(vLVo));
				sb.append(trimVmStdabw(sLVo));
				sb.append(trimV15V85(v15LVo));
				sb.append(trimV15V85(v85LVo));

				if (countLVo == VK_ANZAHL_LV) {
					for (int i = 0; i < countLVo; i++) {
						sb.append(trimQ(vkLVo[i]));
						qLVoVK += vkLVo[i];
					}
				}
				else {
					for (int i = 0; i < VK_ANZAHL_LV; i++) {
						sb.append(trimQ(KEIN_WERT_65535));
					}
				}

				// SGV-Gruppe
				sb.append(trimQ(qSGV));
				sb.append(trimVmStdabw(vSGV));
				sb.append(trimVmStdabw(sSGV));
				sb.append(trimV15V85(v15SGV));
				sb.append(trimV15V85(v85SGV));

				if (countSGV == VK_ANZAHL_SGV) {
					for (int i = 0; i < countSGV; i++) {
						sb.append(trimQ(vkSGV[i]));
						qSGVVK += vkSGV[i];
					}
				}
				else {
					for (int i = 0; i < VK_ANZAHL_SGV; i++) {
						sb.append(trimQ(KEIN_WERT_65535));
					}
				}

				// BPA-Gruppe
				sb.append(trimQ(qBPA));
				sb.append(trimVmStdabw(vBPA));
				sb.append(trimVmStdabw(sBPA));
				sb.append(trimV15V85(v15BPA));
				sb.append(trimV15V85(v85BPA));

				if (countBPA == VK_ANZAHL_BPA) {
					for (int i = 0; i < countBPA; i++) {
						qBPAVK += vkBPA[i];
						sb.append(trimQ(vkBPA[i]));
					}
				}
				else {
					for (int i = 0; i < VK_ANZAHL_BPA; i++) {
						sb.append(trimQ(KEIN_WERT_65535));
					}
				}

				// Prfung Verkehrsmengen - Summe Fahrzeugklassen:
				// Nichtbereinstimmung wie Datenausfall behandeln (siehe javadoc)
				/**       TODO Nach Testphaste wieder auskommentieren.
				 if (qLVo != qLVoVK) {
				 logger.info("Differenz LVo Verkehrsmengen (" + qLVo + ") - Fahrzeugklassen (" + qLVoVK + ")");

				 return "";
				 } else if (qSGV != qSGVVK) {
				 logger.info("Differenz SGV Verkehrsmengen (" + qLVo + ") - Fahrzeugklassen (" + qSGVVK + ")");

				 return "";
				 } else if (qBPA != qBPAVK) {
				 logger.info("Differenz BPA Verkehrsmengen (" + qLVo + ") - Fahrzeugklassen (" + qBPAVK + ")");

				 return "";
				 }
				 **/
			}
			else {

				// Liegen fr eine Stunde keine Daten vor, so bestehen die beiden
				// Datenzeilen lediglich aus Datum, Statuskennung, Stunde und
				// Richtungskennung (Lnge: 14 Zeichen).
				// => damit bleibt der String leer
				return "";
			}
		}

		return sb.toString();
	}

	//~ GET METHODEN ==========================================================

	/**
	 * <p>Liefert die 1. Headerzeile</p> <p>Der erste Header-Datensatz der Stundenwertedatei enthlt eine allgemeine
	 * Beschreibung zur Identifizierung und Lage der Dauerzhlstelle. Die einzelnen Merkmale stehen linksbndig und sind
	 * teilweise durch Leerzeichen voneinander getrennt. Die Satzlnge umfasst genau 67 Stellen. </p>
	 * <pre>
	 * Position Beschreibung                      Beispiel
	 * 1 - 4    Blatt-Nr. der TK25                4807
	 * 5 - 8    Zhlstellennummer                 5033
	 * 9        Leerzeichen
	 * 10 - 11  Bundesland: Kennung               05
	 * 12       Leerzeichen
	 * 13       Straenklasse                     A
	 * 14       Leerzeichen
	 * 15 - 20  Straennummer und ggf. Buchstabe  3
	 * 21       Leerzeichen
	 * 22 - 61  Zhlstellenname                   Opladen II
	 * 62       Leerzeichen
	 * 63 - 66  Versionsnummer der Struktur       V2.0
	 * 67       Zeilenende                        ;
	 * </pre>
	 *
	 * @return {@link String} 1. Headerzeile
	 */
	protected String getHeaderRow01() {
		StringBuilder sb = new StringBuilder(67);

		sb.append(_ZaehlStelle.getTK25AsString());
		sb.append(_ZaehlStelle.getBastNummerAsString()).append(' ');
		sb.append(TWO_DIGIT_FORMAT.format(_ZaehlStelle.getBundesland().getKennung())).append(' ');
		sb.append((_ZaehlStelle.getStrassenKlasse() != null) ? _ZaehlStelle.getStrassenKlasse().getAbbreviation() : ' ').append(' ');

		String s = null;

		if (_ZaehlStelle.getStrasse() != null) {
			s = String.valueOf(_ZaehlStelle.getStrasse().getNummer());

			if (_ZaehlStelle.getStrasse().getZusatz() != null) {
				s = s + _ZaehlStelle.getStrasse().getZusatz();
			}
		}

		sb.append(trim(s, 6)).append(' ');
		sb.append(trim(_ZaehlStelle.getLzzName(), 40)).append(' ');
		sb.append("V2.0;");

		return sb.toString();
	}

	/**
	 * <p>Liefert die 2. Headerzeile</p> <p>Auch der zweite Header-Datensatz der Stundenwertedatei hat eine feste
	 * Satzlnge (168 Zeichen). Der Datensatz enthlt Angaben zur Fahrstreifenanzahl und zu den Fern- und Nahzielen der
	 * beiden Fahrtrichtungen. Die Merkmale, linksbndig und mit Leerzeichen getrennt, sind wie folgt angeordnet: </p>
	 * <pre>
	 * Position  Beschreibung                           Beispiel
	 * 1         Anzahl der Fahrstreifen in Richtung I  3
	 * 2         Leerzeichen
	 * 3         Anzahl der Fahrstreifen in Richtung II 3
	 * 4         Leerzeichen
	 * 5 - 39    Fernziel in Richtung I                 Oberhausen
	 * 40        Leerzeichen
	 * 41 - 85   Nahziel in Richtung I                  AS Solingen
	 * 86        Leerzeichen
	 * 87 - 121  Fernziel in Richtung II                Kln
	 * 122       Leerzeichen
	 * 123 - 167 Nahziel in Richtung II                 AS Opladen
	 * 168       Zeilenende                             ;
	 * </pre>
	 *
	 * @return {@link String} 2. Headerzeile
	 */
	protected String getHeaderRow02() {
		StringBuilder sb = new StringBuilder(168);

		sb.append(_FSCountInRichtung).append(' ');
		sb.append(_FSCountGgRichtung).append(' ');
		sb.append(trim(_ZaehlStelle.getInFernziel(), 35)).append(' ');
		sb.append(trim(_ZaehlStelle.getInNahziel(), 45)).append(' ');
		sb.append(trim(_ZaehlStelle.getGegenFernziel(), 35)).append(' ');
		sb.append(trim(_ZaehlStelle.getGegenNahziel(), 45)).append(';');

		return sb.toString();
	}

	/**
	 * <p>Liefert die 3. Headerzeile.</p> <p>Der dritte Header-Datensatz benennt die Fahrzeuggruppen (FzGr) mit ihren
	 * Bezeichnungen und der Anzahl der Geschwindigkeitsklassen. Danach wird die Richtungskennung und die Fahrzeugmenge
	 * des Schwerverkehrs aufgefhrt. Es folgen die Bezeichnungen fahrzeuggruppenspezifischer Kenngren und die Grenzen
	 * der Geschwindigkeitsklassen der Fahrzeuggruppen. Die Grenze der niedrigsten Geschwindigkeitsklasse liegt bei 0
	 * km/h, und die hchste Geschwindigkeitsklasse ist nach oben hin unbegrenzt. </p> <p>Alle Angaben im dritten
	 * Header-Datensatz sind linksbndig einzutragen. </p> <p>Dauerzhlstellen, die fr die Auswertung von
	 * Geschwindigkeitsdaten herangezogen werden, sollten die Fahrzeugarten zu 3 Fahrzeuggruppen zusammenfassen knnen.
	 * In Ausnahmefllen kann in Abstimmung mit der BASt auf die dritte Fahrzeuggruppe BPA (Busse und Pkw mit Anhnger)
	 * verzichtet werden. </p> <p><i>Hinweis: Im Zusammenspiel mit der TLS Datenversion 24 muss auf keine Fahrzeuggruppe
	 * verzichtet werden. Daher ist das <b>Format fr 2 Fahrzeuggruppen</b> nicht relevant und wird von der Funktion
	 * auch <b>nicht untersttzt</b>!</i> </p>
	 *
	 * @return {@link String} 3. Headerzeile fr Datenformat mit 3 Fahrzeuggruppen
	 */
	protected String getHeaderRow03() {

		// Definition 3. Headerzeile (fr 3 Fz-Gruppen)
		final String HEADER_ROW03_3 = "S3 LVo 16  SGV 10  BPA 10  " + "Ri  qSV qGr vm  svm v15 v85 " + "0   40  50  60  70  80  90  100 110 120 130 140 150 160 170 180 " + "0   40  50  60  70  80  90  100 110 120 " + "0   40  50  60  70  80  90  100 110 120;";

		// Definition 3. Headerzeile (fr 2 Fz-Gruppen/Sonderfall/nicht untersttzt)
		@SuppressWarnings("unused") final String HEADER_ROW03_2 = "S2 LVo 16  SGV 10  " + "Ri  qSV qGr vm  svm v15 v85 " + "0   40  50  60  70  80  90  100 110 120 130 140 150 160 170 180 " + "0   40  50  60  70  80  90  100 110 120;";

		// Momentan wird nur der Standardheader gem TLS Datenversion 24 untersttzt
		return HEADER_ROW03_3;
	}

	/**
	 * <p>Prft, ob das Datum einen gltigen Wert (innerhalb des Wertebereichs liegender Zahlenwert) oder einen
	 * speziellen Zustand hat und gibt danach entweder den skalierten Zahlenwert als <code>double</code> oder
	 * <code>Double.NaN</code> im Falle eines Zustands zurck.</p> <p>Die Funktion ist bei Lesen von
	 * Geschwindigkeitswerten zu verwenden, bei denen es laut TLS den Zustand "Wert nicht ermittelbar" geben kann. Beim
	 * Datentyp 79 triff dies auf folgende Attribute zu:<p> <ul> <li>mittlere Geschwindigkeit (0..6553,4 km/h,
	 * 65535="Wert nicht ermittelbar")</li> <li>Standardabweichung (0..6553,4 km/h, 65535="Wert nicht ermittelbar")</li>
	 * <li>v85 Geschwindigkeit (0..254 km/h, 255="Wert nicht ermittelbar")</li> <li>v15 Geschwindigkeit (0..254 km/h,
	 * 255="Wert nicht ermittelbar")</li> </ul>
	 *
	 * @param data     Data.
	 * @param itemName Item Name.
	 *
	 * @return Double.NaN wenn kein gltiger Wert vorliegt oder sich der skalierte Wert nicht innerhalb des
	 *         Wertebereichs befindet, ansonsten der skalierte Wert als double
	 */
	protected double getScaledValueAsDouble(Data data, String itemName) {
		double result = Double.NaN;

		if (data != null) {
			Data.NumberValue item = data.getScaledValue(itemName);

			if (item != null) {
				if (!item.isState() && item.isNumber()) {
					result = item.doubleValue();
				}
			}
		}

		return result;
	}

	//~ METHODEN ==============================================================

	/**
	 * <p>Erzeugt die BASt-Format-Daten und spooled diese.</p>
	 * <p/>
	 * <p>Die bergebenen Daten mssen vom Typ atg.tlsLveErgebnisMeldungVersion24 sein und vor dem ersten Aufruf muss
	 * der Channel vollstndig initialisiert werden. </p> <p>Die BASt-Datei beginnt mit 3 Headerzeilen.</p> <p>Nach dem
	 * dritten Header-Datensatz folgen die einzelnen Stundendatenstze. Dabei werden fr jede einzelne Stunde genau zwei
	 * Datenstze (Zeilen) erzeugt, fr jede Fahrtrichtung einen. Die Lnge der Datenstze ist von der Anzahl der
	 * Fahrstreifen und der Fahrzeuggruppen abhngig.<br> Die ersten 6 Positionen enthalten das Datum (Pos. 1-6, Format:
	 * yymmdd). Nach einer Statuskennung  fr MESZ/MEZ  folgt die Stunde (Pos. 8-12, Format: hh:mm, beginnend mit
	 * 01:00). Ein Leerzeichen an Position 13 trennt die Richtungskennung (1 bzw. 2) ab. Alle weiteren Angaben
	 * sind ebenfalls durch Leerzeichen voneinander getrennt.<br> Es folgen fr den ersten (rechten) Fahrstreifen der
	 * Richtung die Fahrzeugmenge des Schwerverkehrs qSV und anschlieend fahrzeuggruppenbezogen die Fahrzeugmenge qGr
	 * der Fahrzeuggruppe 1, die mittlere Geschwindigkeit vm dieser Fahrzeuggruppe auf eine Nachkommastelle gerundet mit
	 * , als Trennzeichen (nicht mit dem amerikanischen Dezimalzeichen .) und die zugehrige Standardabweichung svm,
	 * ebenfalls auf eine Nachkommastelle gerundet mit , als Trennzeichen sowie die v15 und v85. Anschlieend werden
	 * die Verkehrsmengen jeder Geschwindigkeitsklasse der ersten Fahrzeuggruppe angegeben.<br> Eine analoge Auflistung
	 * fr den rechten Fahrstreifen erfolgt fr die Fahrzeuggruppe 2 und, falls eine dritte Fahrzeuggruppe erfasst wird,
	 * auch fr sie. Die Werte fr weitere Fahrstreifen dieser Richtung sind von auen zur Fahrbahnmitte anzugeben.<br>
	 * Liegen fr eine Stunde keine Daten vor, so bestehen die beiden Datenzeilen lediglich aus Datum, Statuskennung,
	 * Stunde und Richtungskennung (Lnge: 14 Zeichen). [Anmerkung: auch fehlende Fahrzeugraster werden wie
	 * Datenausflle behandelt, siehe {@link #DataChannelBAStVV2007}].<br> Die Angaben in den Stundendatenstzen sind im
	 * Gegensatz zu den drei Header- Datenstzen rechtsbndig ausgerichtet. </p>
	 *
	 * @param rsArrayAtTime ResultData[]
	 * @param calendar      Intervallbeginn der Datens&auuml;
	 *
	 * @throws ChannelException
	 */
	public void processRecord(Dataset[] rsArrayAtTime, Calendar calendar) throws ChannelException {
		int actHour = calendar.get(Calendar.HOUR);

		// Prfen, ob Umstellung MEZ auf MESZ (Sommerzeit), dann gibt es fr den
		// letzten Sonntag im Mrz 2 Uhr keine Daten und leere Zeilen mit der
		// Kennung 'm' mssen eingefgt werden.
		if ((_LastHour == 1) && (actHour == 3)) {
			for (int ri = 1; ri <= 2; ri++) {
				StringBuilder sb = new StringBuilder();

				sb.append(DATE_FORMAT.format(calendar.getTime())).append("m03:00 ").append(ri);
				this.spool(sb);
			}
		}

		// Prfung, ob Umstellung MESZ auf MEZ (Winterzeit), dann gibt es fr die
		// 2 Uhr Stunde am letzten Sonntag im Oktober doppelt und die Werte
		// beider Stunden sind zusammenzufassen und mit 'o' zu kennzeichnen.
		if (_LastHour == actHour) {
			this.deleteLastLine();  // Richtung 2
			this.deleteLastLine();  // Richtung 1

			for (int ri = 1; ri <= 2; ri++) {
				StringBuilder sb = new StringBuilder();

				sb.append(DATE_FORMAT.format(calendar.getTime())).append("o03:00 ").append(ri);
				sb.append(extractData(rsArrayAtTime, _LastArrayAtTime, (ri == 1)));
				this.spool(sb);
			}
		}
		else {

			// normale Stunde
			for (int ri = 1; ri <= 2; ri++) {
				StringBuilder sb = new StringBuilder();

				sb.append(DATE_FORMAT.format(calendar.getTime())).append(' ').append(TWO_DIGIT_FORMAT.format(calendar.get(Calendar.HOUR_OF_DAY) + 1)).append(":00").append(' ').append(ri);
				sb.append(extractData(rsArrayAtTime, null, (ri == 1)));
				this.spool(sb);
			}
		}

		_LastHour = actHour;
		_LastArrayAtTime = rsArrayAtTime;
	}

	//~ SET METHODEN ==========================================================

	/**
	 * <p>Setzt die zu verarbeitende Zhlstelle und initialisiert die zur Verarbeitung bentigten Header- und
	 * Richtungsangaben.</p>
	 *
	 * @see de.inovat.kex.bast.exporter.AbstractDataChannelBASt#setZaehlstelle(de.inovat.kex.bast.common.objectwrapper.LangZeitZaehlStelle)
	 */
	public void setZaehlstelle(LangZeitZaehlStelle zaehlstelle) throws ChannelException {
		_ZaehlStelle = zaehlstelle;
		if (_ZaehlStelle != null) {

			_FSCountInRichtung = zaehlstelle.getInAuswertequerschnitt().getMessQuerschnitt().getFahrstreifen().length;
			_FSCountGgRichtung = zaehlstelle.getGegenAuswertequerschnitt().getMessQuerschnitt().getFahrstreifen().length;
			_CountOfFS = _FSCountInRichtung + _FSCountGgRichtung;
		}
	}

	//~ METHODEN ==============================================================

	/**
	 * <p>Formatiert den String linksbndig und fllt bis zur angegebenen Lnge mit ' ' auf. Ist der String
	 * <code>null</code>, wird vollstndig mit Leerzeichen aufgefllt. berzhlige Zeichen werden abgeschnitten.</p>
	 *
	 * @param string auszugebender String, <code>null</code> ist erlaubt und wird als "" interpretiert
	 * @param length Lnge bis zu der mit ' ' aufgefllt wird oder nach der der String abgeschnitten wird
	 *
	 * @return linksbndig ausgerichteter String der Lnge <code>length</code>
	 */
	protected String trim(String string, int length) {
		return String.format("%-" + length + "s", (string != null) ? string : "");

		// an dieser Stelle sind keine nennenswerten Optimierungen mglich, bei
		// Verwendung von format(...) dauert die Textausgabe <1s / Monatsdatei
	}

	/**
	 * <p>Formatiert die Fahrzeugmenge als rechtsbndigen String.</p> <p>Gem Vorgabe sind die Angaben in den
	 * Stundendatenstzen im Gegensatz zu den drei Header- Datenstzen rechtsbndig ausgerichtet.<br> Bei Sonderwerten
	 * der TLS fr nicht ermittelbare bzw. fehlerhafte Werte, wird "     0" in die Ausgabedatei geschrieben. Das
	 * betrifft negative Werte oder alle >=65535.</p>
	 *
	 * @param value auszugebender Wert (Fahrzeuganzahl, die Werte mssen zwischen 0 und 65534 liegen, Wert 65535 =
	 *              Zustand "nicht ermittelbar")
	 *
	 * @return String mit dem rechtsbndig formatierten Wert
	 */
	protected String trimQ(int value) {
		if ((value >= 0) && (value < KEIN_WERT_65535)) {
			return String.format(Locale.GERMANY, "%6d", value);

			// an dieser Stelle sind keine nennenswerten Optimierungen mglich, bei
			// Verwendung von format(...) dauert die Textausgabe <1s / Monatsdatei
		}
		else {

			/** @TODO IllegalArgumentException ? */
			return "     0";
		}
	}

	/**
	 * <p>Formatiert die v15 und v85 Geschwindigkeit rechtsbndig ganzzahlig.</p> <p>Gem Vorgabe sind die v15 und v85
	 * (entsprechend der Erfassung) als ganzzahligen Wert.</p> <p>Bei Sonderwerten der TLS fr nicht ermittelbare bzw.
	 * fehlerhafte Werte, wird "     0" in die Ausgabedatei geschrieben. Das betrifft negative Werte und alle >=255
	 * </p>
	 *
	 * @param value auszugebender Wert (Geschwindigkeit, die Werte mssen zwischen 0 und 254 liegen, Wert 255="nicht
	 *              ermittelbar)
	 *
	 * @return String mit rechtsbndig formatiertem Wert
	 */
	protected String trimV15V85(double value) {
		if ((value >= 0) && (value < KEIN_WERT_255)) {
			return String.format(Locale.GERMANY, "%6.0f", value);

			// an dieser Stelle sind keine nennenswerten Optimierungen mglich, bei
			// Verwendung von format(...) dauert die Textausgabe <1s / Monatsdatei
		}
		else {

			/** @TODO IllegalArgumentException ? */
			return "     0";
		}
	}

	/**
	 * <p>Formatiert die mittlere Geschwindigkeit und deren Standardabweichung rechtsbndig mit einer
	 * Nachkommastelle.</p> <p>Gem Vorgabe ist die mittlere Geschwindigkeit vm auf eine Nachkommastelle zu runden mit
	 * ',' als Trennzeichen (nicht mit dem amerikanischen Dezimalzeichen '.) und die zugehrige Standardabweichung svm
	 * ebenfalls auf eine Nachkommastelle gerundet mit ',' als Trennzeichen. </p> <p>Bei Sonderwerten der TLS fr nicht
	 * ermittelbare bzw. fehlerhafte Werte, wird "   0,0" in die Ausgabedatei geschrieben. Das betrifft negative Werte
	 * und alle >=6553.5.</p>
	 *
	 * @param value auszugebender Wert (Geschwindigkeit, die Werte mssen zwischen 0 und 6553,5 liegen, Wert 6553.5 =
	 *              Zustand "nicht ermittelbar")
	 *
	 * @return String mit rechtsbndig formatiertem Wert
	 */
	protected String trimVmStdabw(double value) {
		if ((value >= 0) && (value < KEIN_WERT_6553_5)) {
			return String.format(Locale.GERMANY, "%6.1f", value);

			// an dieser Stelle sind keine nennenswerten Optimierungen mglich, bei
			// Verwendung von format(...) dauert die Textausgabe <1s / Monatsdatei
		}
		else {

			/** @TODO IllegalArgumentException ? */
			return "   0,0";
		}
	}

	/**
	 * <p>Schreibt die Headerzeilen 1, 2 und 3 in die Datei.</p>
	 *
	 * @param writer PrintWriter.
	 */
	protected void writeHeader(PrintWriter writer) {
		writer.println(getHeaderRow01());
		writer.println(getHeaderRow02());
		writer.println(getHeaderRow03());
	}

	@Override
	protected void writeHeaderFTP(BufferedOutputStream bos) throws IOException {
		bos.write(getHeaderRow01().getBytes());
		bos.write(System.getProperty("line.separator").getBytes());
		bos.write(getHeaderRow02().getBytes());
		bos.write(System.getProperty("line.separator").getBytes());
		bos.write(getHeaderRow03().getBytes());
		bos.write(System.getProperty("line.separator").getBytes());
	}
}

//~Formatiert mit 'inovat Kodierkonvention' am 21.05.10
